__version__ = "0.1.1"

__title__ = "pytation"
__description__ = 'Create manufacturing and validation tests for hardware development.'
__url__ = 'https://github.com/jetperch/pytation'
__author__ = 'Jetperch LLC'
__author_email__ = 'joulescope-dev@jetperch.com'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2018-2021 Jetperch LLC'

__all__ = ['__version__', '__title__', '__description__', '__url__',
           '__author__', '__author_email__', '__license__', '__copyright__']
